from inspect import isclass
from enum import Enum

from sqlalchemy import types

from pulsar import ImproperlyConfigured


class ScalarCoercible(object):
    def _coerce(self, value):
        raise NotImplementedError

    def coercion_listener(self, target, value, oldvalue, initiator):
        return self._coerce(value)


class Choice(object):
    def __init__(self, code, value):
        self.code = code
        self.value = value

    def __eq__(self, other):
        if isinstance(other, Choice):
            return self.code == other.code
        return other == self.code

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return 'Choice(code={code}, value={value})'.format(
            code=self.code,
            value=self.value
        )


class ChoiceType(types.TypeDecorator, ScalarCoercible):
    impl = types.Unicode(255)

    def __init__(self, choices, impl=None, **kwargs):
        self.choices = choices

        if isinstance(choices, type) and issubclass(choices, Enum):
            self.type_impl = EnumTypeImpl(enum_class=choices, **kwargs)
        else:
            self.type_impl = ChoiceTypeImpl(choices=choices, **kwargs)

        if impl:
            if isclass(impl):
                impl = impl()
            self.impl = impl

    @property
    def python_type(self):
        return self.impl.python_type

    def _coerce(self, value):
        return self.type_impl._coerce(value)

    def process_bind_param(self, value, dialect):
        return self.type_impl.process_bind_param(value, dialect)

    def process_result_value(self, value, dialect):
        return self.type_impl.process_result_value(value, dialect)


class ChoiceTypeImpl(object):
    """The implementation for the ``Choice`` usage."""

    def __init__(self, choices):
        if not choices:
            raise ImproperlyConfigured(
                'ChoiceType needs list of choices defined.'
            )
        self.choices_dict = dict(choices)

    def _coerce(self, value):
        if value is None:
            return value
        if isinstance(value, Choice):
            return value
        return Choice(value, self.choices_dict[value])

    def process_bind_param(self, value, dialect):
        if value and isinstance(value, Choice):
            return value.code
        return value

    def process_result_value(self, value, dialect):
        if value:
            return Choice(value, self.choices_dict[value])
        return value


class EnumTypeImpl(object):
    """The implementation for the ``Enum`` usage."""

    def __init__(self, enum_class, bind_by_name=True):
        self.enum_class = enum_class
        self.bind_by_name = bind_by_name

    def _coerce(self, value):
        return self.enum_class(value) if value else None

    def process_bind_param(self, value, dialect):
        ret = None
        if isinstance(value, Enum):
            ret = value.value
        elif self.bind_by_name and isinstance(value, str):
            for e in self.enum_class:
                if e.name.lower() == value.lower():
                    ret = e.value
        elif value:
            ret = self.enum_class(value).value
        return ret

    def process_result_value(self, value, dialect):
        return self.enum_class(value) if value else None



import logging
import sys
from copy import copy
from inspect import getmodule
from contextlib import contextmanager
from collections import OrderedDict

from sqlalchemy import MetaData, Table, event
from sqlalchemy.ext.declarative.api import (declarative_base, declared_attr,
                                            _as_declarative, _add_attribute)
from sqlalchemy.orm.session import Session
from sqlalchemy.orm import object_session
from sqlalchemy.schema import DDL

from pulsar import ImproperlyConfigured

from .strategy import create_engine
from .utils import database_operation


logger = logging.getLogger('pulsar.odm')


class OdmMeta(type):

    def __new__(cls, name, bases, attrs):
        abstract = attrs.pop('__odm_abstract__', False)
        klass = super().__new__(cls, name, bases, attrs)
        if not abstract and not isinstance(klass, DeclarativeMeta):
            module = getmodule(klass)
            models = getattr(module, '__odm_models__', None)
            if models is None:
                models = OrderedDict()
                module.__odm_models__ = models
            name = klass.__name__.lower()
            models[name] = klass
        return klass


class DeclarativeMeta(OdmMeta):

    def __init__(cls, classname, bases, dict_):
        if '_decl_class_registry' not in cls.__dict__:
            _as_declarative(cls, classname, cls.__dict__)
        type.__init__(cls, classname, bases, dict_)

    def __setattr__(cls, key, value):
        _add_attribute(cls, key, value)


class BaseModel(metaclass=OdmMeta):
    __odm_abstract__ = True

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    @classmethod
    def create_table(cls, name, *columns, **kwargs):
        targs = table_args(cls, **kwargs)
        args, kwargs = targs[:-1], targs[-1]
        table = Table(name, MetaData(), *columns, *args, **kwargs)
        return table


def table_args(*args, **kwargs):
    targs = ()
    tkwargs = {}

    if args:
        if hasattr(args[0], '__table_args__'):
            targs = args[0].__table_args__
            targs, tkwargs = targs[:-1], targs[-1].copy()
            args = args[1:]

        targs += args

    for key, value in kwargs.items():
        if isinstance(value, dict) and key in tkwargs:
            new_value = tkwargs[key].copy()
            new_value.update(value)
            value = new_value
        tkwargs[key] = value

    return targs + (tkwargs,)


def model_base(bind_label=None, info=None):
    """Create a base declarative class
    """
    Model = type('Model', (BaseModel,), {'__odm_abstract__': True})
    info = {}
    Model.__table_args__ = table_args(info=info)
    if bind_label:
        info['bind_label'] = bind_label
    return Model


def module_tables(module):
    for name, table in vars(module).items():
        if isinstance(table, Table):
            yield table


def copy_models(module_from, module_to):
    """Copy models from one module to another
    :param module_from:
    :param module_to:
    :return:
    """
    module_from = get_module(module_from)
    module_to = get_module(module_to)
    models = get_models(module_from)
    if models:
        models = models.copy()
        models.update(((t.key, t) for t in module_tables(module_from)))
        module_to.__odm_models__ = models
        return models


def move_models(module_from, module_to):
    module_from = get_module(module_from)
    if copy_models(module_from, module_to):
        del module_from.__odm_models__


def get_module(module_or_name):
    if isinstance(module_or_name, str):
        return sys.modules[module_or_name]
    else:
        return getmodule(module_or_name)


def get_models(module):
    """Get models from a module
    :param module:
    :return:
    """
    return getattr(get_module(module), '__odm_models__', None)


Model = model_base()


class Mapper:
    """SQLAlchemy wrapper

    .. attribute:: binds

        Dictionary of labels-engine pairs. The "default" label is always
        present and it is used for tables without `bind_label` in their
        `info` dictionary.
    """
    def __init__(self, binds):
        # Setup mdoels and engines
        if not binds:
            binds = {}
        elif isinstance(binds, str):
            binds = {'default': binds}
        if 'default' not in binds:
            raise ImproperlyConfigured('default datastore not specified')

        self._engines = {}
        self._declarative_register = {}
        self._bases = {}
        self._base_declarative = declarative_base(name='OdmBase',
                                                  metaclass=DeclarativeMeta)
        self.binds = {}
        self.is_green = False

        for name, bind in tuple(binds.items()):
            key = None if name == 'default' else name
            engine = create_engine(bind)
            dialect = engine.dialect
            # Dialect requires Green Pool
            if getattr(dialect, 'is_green', False):
                self.is_green = True
            self._engines[key] = engine

    def __getitem__(self, model):
        return self._declarative_register[model]

    def __getattr__(self, name):
        if name in self._declarative_register:
            return self._declarative_register[name]
        raise AttributeError('No model named "%s"' % name)

    @property
    def metadata(self):
        """Returns the :class:`~sqlalchemy.Metadata` for this mapper
        """
        return self._base_declarative.metadata

    def copy(self, binds):
        return self.__class__(binds)

    def register(self, model):
        """Register a model or a table with this mapper

        :param model: a table or a :class:`.BaseModel` class
        :return: a Model class or a table
        """
        metadata = self.metadata
        if not isinstance(model, Table):
            model_name = self._create_model(model)
            if not model_name:
                return
            model, name = model_name
            table = model.__table__
            self._declarative_register[name] = model

            if name in self._bases:
                for model in self._bases.pop(name):
                    self.register(model)
        else:
            table = model.tometadata(metadata)
            model = table

        # Register engine
        engine = None
        label = table.info.get('bind_label')
        keys = ('%s.%s' % (label, table.key),
                label, None) if label else (None,)
        #
        # Find the engine for this table
        for key in keys:
            engine = self.get_engine(key)
            if engine:
                break
        assert engine
        self.binds[table] = engine

        return model

    def register_module(self, module, exclude=None):
        module = get_module(module)
        models = get_models(module)
        exclude = set(exclude or ())
        if models:
            for name, model in models.items():
                if name in exclude:
                    continue
                self.register(model)
        for table in module_tables(module):
            if table.key not in exclude:
                self.register(table)

    def create_table(self, name, *columns, **kwargs):
        """Create a new table with the same metadata and info
        """
        targs = table_args(**kwargs)
        args, kwargs = targs[:-1], targs[-1]
        return Table(name, self.metadata, *columns, *args, **kwargs)

    def database_create(self, database, **params):
        """Create databases for each engine and return a new :class:`.Mapper`.
        """
        binds = {}
        dbname = database
        for key, engine in self.keys_engines():
            if hasattr(database, '__call__'):
                dbname = database(engine)
            assert dbname, "Cannot create a database, no db name given"
            key = key if key else 'default'
            binds[key] = self._database_create(engine, dbname)
        return self.copy(binds)

    def database_exist(self):
        """Create databases for each engine and return a new :class:`.Mapper`.
        """
        binds = {}
        for key, engine in self.keys_engines():
            key = key if key else 'default'
            binds[key] = self._database_exist(engine)
        return binds

    def database_all(self):
        """Return a dictionary mapping engines with databases
        """
        all = {}
        for engine in self.engines():
            all[engine] = self._database_all(engine)
        return all

    def database_drop(self, database=None, **params):
        dbname = database
        for engine in self.engines():
            if hasattr(database, '__call__'):
                dbname = database(engine)
            assert dbname, "Cannot drop database, no db name given"
            self._database_drop(engine, dbname)

    def tables(self):
        tables = []
        for engine in self.engines():
            tbs = engine.table_names()
            if tbs:
                tables.append((str(engine.url), tbs))
        return tables

    def table_create(self, remove_existing=False):
        """Creates all tables.
        """
        for engine in self.engines():
            tables = self._get_tables(engine, create_drop=True)
            logger.info('Create all tables for %s', engine)
            try:
                self.metadata.create_all(engine, tables=tables)
            except Exception as exc:
                raise

    def table_drop(self):
        """Drops all tables.
        """
        for engine in self.engines():
            tables = self._get_tables(engine, create_drop=True)
            logger.info('Drop all tables for %s', engine)
            self.metadata.drop_all(engine, tables=tables)

    @contextmanager
    def begin(self, close=True, expire_on_commit=False, session=None,
              commit=False, **options):
        """Provide a transactional scope around a series of operations.

        By default, ``expire_on_commit`` is set to False so that instances
        can be used outside the session.
        """
        if not session:
            commit = True
            session = self.session(expire_on_commit=expire_on_commit,
                                   **options)
        else:
            close = False
        try:
            yield session
            if commit:
                session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            if close:
                session.close()

    def session(self, **options):
        options['binds'] = self.binds
        return OdmSession(self, **options)

    def session_from_object(self, *objs):
        for obj in objs:
            session = object_session(obj)
            if session is not None:
                return session

    def dialect(self, key):
        """Dialect object for a model/table name
        """
        return self.binds[self[key].__table__].dialect

    def get_engine(self, key=None):
        """Get an engine by key
        """
        if key in self._engines:
            return self._engines[key]

    def engines(self):
        """Iterator over all engines
        """
        return self._engines.values()

    def keys_engines(self):
        return self._engines.items()

    def close(self):
        for engine in self.engines():
            engine.dispose()

    # INTERNALS
    def _create_model(self, model):
        model_name = model.__name__
        meta = type(self._base_declarative)
        if isinstance(model, meta):
            raise ImproperlyConfigured('Cannot register declarative classes '
                                       'only mixins allowed')
        base = getattr(model, '__inherit_from__', None)
        if base:
            if base not in self._declarative_register:
                models = self._bases.get(base)
                if not models:
                    self._bases[base] = models = []
                models.append(model)
                return
            else:
                base = self._declarative_register[base]
        else:
            base = self._base_declarative

        #
        # Create SqlAlchemy Model
        model = meta(model_name, (model, base), {})
        create = getattr(model, '__create_sql__', None)
        name = model_name.lower()
        if create:
            event.listen(self.metadata,
                         'after_create',
                         DDL(create.format({'name': name})))
            drop = getattr(model, '__drop_sql__', None)
            if not drop:
                logger.warning('Model %s has create statement but not drop. '
                               'To mute this warning add a __drop_sql__ '
                               'statement in the model class', name)
            else:
                event.listen(self.metadata,
                             'before_drop',
                             DDL(drop.format({'name': name})))

        return model, name

    def _get_tables(self, engine, create_drop=False):
        tables = []
        for table, eng in self.binds.items():
            if eng == engine:
                if table.key in self._declarative_register:
                    model = self[table.key]
                    if create_drop and hasattr(model, '__create_sql__'):
                        continue
                tables.append(table)
        return tables

    def _database_all(self, engine):
        return database_operation(engine, 'all')

    def _database_create(self, engine, database):
        """Create a new database and return a new url representing
        a connection to the new database
        """
        logger.info('Creating database "%s" in "%s"', database, engine)
        database_operation(engine, 'create', database)
        url = copy(engine.url)
        url.database = database
        return str(url)

    def _database_drop(self, engine, database):
        logger.info('dropping database "%s" from "%s"', database, engine)
        database_operation(engine, 'drop', database)

    def _database_exist(self, engine):
        return database_operation(engine, 'exists')


class OdmSession(Session):

    def __init__(self, mapper, **options):
        self.mapper = mapper
        super().__init__(**options)



"""JSONType definition."""
import json

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSON, JSONB


class JSONType(sa.types.TypeDecorator):
    """
    JSONType offers way of saving JSON data structures to database. On
    PostgreSQL the underlying implementation of this data type is 'json' while
    on other databases its simply 'text'.

    ::


        from odm.types import JSONType


        class Product(Base):
            __tablename__ = 'product'
            id = sa.Column(sa.Integer, autoincrement=True)
            name = sa.Column(sa.Unicode(50))
            details = sa.Column(JSONType)


        product = Product()
        product.details = {
            'color': 'red',
            'type': 'car',
            'max-speed': '400 mph'
        }
        session.commit()
    """
    impl = sa.UnicodeText

    def __init__(self, binary=True, impl=sa.UnicodeText, *args, **kwargs):
        self.binary = binary
        self.impl = impl
        super().__init__(*args, **kwargs)

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            # Use the native JSONB or JSON type.
            if self.binary:
                return dialect.type_descriptor(JSONB)
            else:
                return dialect.type_descriptor(JSON)
        else:
            return dialect.type_descriptor(self.impl)

    def process_bind_param(self, value, dialect):
        if dialect.name == 'postgresql':
            return value
        elif value is not None:
            return json.dumps(value)
        else:
            return value

    def process_result_value(self, value, dialect):
        if dialect.name == 'postgresql':
            return value
        elif value is not None:
            return json.loads(value)
        else:
            return value



import sqlalchemy
from sqlalchemy.engine import url
from sqlalchemy.engine.strategies import PlainEngineStrategy


def create_engine(*args, **kwargs):
    kwargs.setdefault('strategy', 'odm')
    return sqlalchemy.create_engine(*args, **kwargs)


class OdmEngineStrategy(PlainEngineStrategy):
    name = 'odm'

    def create(self, name_or_url, **kwargs):
        # create url.URL object
        u = url.make_url(name_or_url)

        if 'pool_size' in u.query:
            kwargs['pool_size'] = int(u.query.pop('pool_size'))
        if 'pool_timeout' in u.query:
            kwargs['pool_timeout'] = float(u.query.pop('pool_timeout'))

        return super().create(name_or_url, **kwargs)


OdmEngineStrategy()



from .choice import ChoiceType
from .ip_address import IPAddressType
from .json import JSONType
from .uuid import UUIDType


__all__ = ('ChoiceType',
           'IPAddressType',
           'JSONType',
           'UUIDType')



from asyncio import Future

from greenlet import getcurrent
import psycopg2
from psycopg2 import *  # noqa
from psycopg2 import extensions, OperationalError


__version__ = psycopg2.__version__


def psycopg2_wait_callback(conn):
    """A wait callback to allow greenlet to work with Psycopg.
    The caller must be from a greenlet other than the main one.

    :param conn: psycopg2 connection or file number

    This function must be invoked from a coroutine with parent, therefore
    invoking it from the main greenlet will raise an exception.
    """
    while True:
        state = conn.poll()
        if state == extensions.POLL_OK:
            # Done with waiting
            break
        elif state == extensions.POLL_READ:
            _wait_fd(conn)
        elif state == extensions.POLL_WRITE:
            _wait_fd(conn, read=False)
        else:  # pragma    nocover
            raise OperationalError("Bad result from poll: %r" % state)


# INTERNALS

def _wait_fd(conn, read=True):
    '''Wait for an event on file descriptor ``fd``.

    :param conn: file descriptor
    :param read: wait for a read event if ``True``, otherwise a wait
        for write event.

    This function must be invoked from a coroutine with parent, therefore
    invoking it from the main greenlet will raise an exception.
    '''
    current = getcurrent()
    parent = current.parent
    assert parent, '"_wait_fd" must be called by greenlet with a parent'
    try:
        fileno = conn.fileno()
    except AttributeError:
        fileno = conn
    future = Future()
    # When the event on fd occurs switch back to the current greenlet
    if read:
        future._loop.add_reader(fileno, _done_wait_fd, fileno, future, read)
    else:
        future._loop.add_writer(fileno, _done_wait_fd, fileno, future, read)
    # switch back to parent greenlet
    parent.switch(future)
    # Back on the child greenlet. Raise error if there is one
    future.result()


def _done_wait_fd(fd, future, read):
    try:
        if read:
            future._loop.remove_reader(fd)
        else:
            future._loop.remove_writer(fd)
    except Exception as exc:
        future.set_exception(exc)
    else:
        future.set_result(None)


try:
    extensions.POLL_OK
except AttributeError:  # pragma    nocover
    from pulsar import ImproperlyConfigured
    raise ImproperlyConfigured(
        'Psycopg2 does not have support for asynchronous connections. '
        'You need at least version 2.2.0 of Psycopg2.')

extensions.set_wait_callback(psycopg2_wait_callback)



import os
from inspect import isclass

import sqlalchemy as sa
from sqlalchemy import inspect
from sqlalchemy.exc import OperationalError, ProgrammingError


def get_columns(mixed):
    """
    Return a collection of all Column objects for given SQLAlchemy
    object.
    The type of the collection depends on the type of the object to return the
    columns from.
    ::
        get_columns(User)
        get_columns(User())
        get_columns(User.__table__)
        get_columns(User.__mapper__)
        get_columns(sa.orm.aliased(User))
        get_columns(sa.orm.alised(User.__table__))

    :param mixed:
        SA Table object, SA Mapper, SA declarative class, SA declarative class
        instance or an alias of any of these objects
    """
    if isinstance(mixed, sa.Table):
        return mixed.c
    if isinstance(mixed, sa.orm.util.AliasedClass):
        return sa.inspect(mixed).mapper.columns
    if isinstance(mixed, sa.sql.selectable.Alias):
        return mixed.c
    if isinstance(mixed, sa.orm.Mapper):
        return mixed.columns
    if not isclass(mixed):
        mixed = mixed.__class__
    return sa.inspect(mixed).columns


def database_operation(engine, oper, *args):
    operation = _database_operation(engine, oper)
    return operation(engine, *args)


def _database_operation(engine, oper):
    dialect = engine.dialect
    method_name = 'database_%s' % oper
    if hasattr(dialect, method_name):
        return getattr(dialect, method_name)
    else:
        scripts = engine_scripts[method_name]
        if hasattr(scripts, dialect.name):
            return getattr(scripts, dialect.name)
        else:
            return scripts.default


class CreateDatabase:

    def sqlite(self, engine, database):
        pass

    def default(self, engine, database):
        conn = engine.connect()
        # the connection will still be inside a transaction,
        # so we have to end the open transaction with a commit
        conn.execute("commit")
        conn.execute('create database %s' % database)
        conn.close()


class DropDatabase:

    def sqlite(self, engine, database):
        try:
            os.remove(database)
        except FileNotFoundError:
            pass

    def default(self, engine, database):
        conn = engine.connect()
        conn.execute("commit")
        conn.execute('drop database %s' % database)
        conn.close()


class AllDatabase:

    def sqlite(self, engine):
        database = engine.url.database
        if os.path.isfile(database):
            return [database]
        else:
            return []

    def default(self, engine):
        insp = inspect(engine)
        return insp.get_schema_names()


class ExistDatabase:
    """Check if a database exists"""

    def sqlite(self, engine):
        database = engine.url.database
        if database:
            return database == ':memory:' or os.path.exists(database)
        else:
            # The default SQLAlchemy database is in memory,
            # and :memory is not required, thus we should
            # support that use-case
            return True

    def postgresql(self, engine):
        database = engine.url.database
        text = "SELECT 1 FROM pg_database WHERE datname='%s'" % database
        return bool(engine.execute(text).scalar())

    def mysql(self, engine):
        database = engine.url.database
        text = ("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
                "WHERE SCHEMA_NAME = '%s'" % database)
        return bool(engine.execute(text).scalar())

    def default(self, engine):
        try:
            engine.execute('SELECT 1')
            return True
        except (ProgrammingError, OperationalError):
            return False


engine_scripts = {'database_exists': ExistDatabase(),
                  'database_create': CreateDatabase(),
                  'database_drop': DropDatabase(),
                  'database_all': AllDatabase()}



from sqlalchemy import pool

from pulsar.apps.greenio import GreenLock


class GreenletPool(pool.Pool):
    '''A Pool that maintains one connection per greenlet.

    Maintains one connection per each greenlet, never moving a
    connection to a greenlet other than the one which it was
    created in.
    '''
    def __init__(self, creator, pool_size=10, timeout=30, **kw):
        super().__init__(creator, **kw)
        self._lock = GreenLock()
        self._max_size = pool_size
        self._timeout = timeout
        self._connections = set()
        self._available_connections = set()

    def dispose(self):
        for conn in self._connections:
            try:
                conn.close()
            except Exception:
                # pysqlite won't even let you close a conn from a thread
                # that didn't create it
                pass
        self.logger.info("Pool disposed. %s", self.status())

    def status(self):
        return "size: %d, available: %d" % (self.size(),
                                            len(self._available_connections))

    def size(self):
        return len(self._connections)

    def max_size(self):
        return self._max_size

    def timeout(self):
        return self._timeout

    def recreate(self):
        self.logger.info("Pool recreating")
        return self.__class__(self._creator,
                              pool_size=self.max_size,
                              recycle=self._recycle,
                              echo=self.echo,
                              logging_name=self._orig_logging_name,
                              use_threadlocal=self._use_threadlocal,
                              reset_on_return=self._reset_on_return,
                              _dispatch=self.dispatch,
                              dialect=self._dialect)

    def _do_return_conn(self, conn):
        self._available_connections.add(conn)

    def _do_get(self):
        try:
            return self._available_connections.pop()
        except KeyError:
            pass

        # Only create one connection at a time, otherwise psycopg2 block!
        with self._lock:
            conn = self._create_connection()
            self._connections.add(conn)
            return conn



from sqlalchemy.dialects.postgresql.psycopg2 import PGDialect_psycopg2
from sqlalchemy.dialects import registry

from .pool import GreenletPool


class PGDGreen(PGDialect_psycopg2):
    '''PostgreSql dialect using psycopg2 and greenlet to obtain an
    implicit asynchronous database connection.
    '''
    poolclass = GreenletPool
    is_green = True

    @classmethod
    def dbapi(cls):
        from odm.dialects.postgresql import green
        return green

    def create_connect_args(self, url):
        args, opts = super().create_connect_args(url)
        opts.pop('pool_size', None)
        opts.pop('pool_timeout', None)
        return [[], opts]


registry.register("postgresql.green", "odm.dialects.postgresql", "PGDGreen")



'''Object Data Mapper for pulsar asynchronous framework'''
import os
from .version import get_version


VERSION = (0, 6, 2, 'final', 0)
__version__ = get_version(VERSION, __file__)


if os.environ.get('package_info') != 'odm':

    from .mapper import (Model, Mapper, OdmSession, logger, model_base,
                         BaseModel, table_args, declared_attr, copy_models,
                         move_models, get_models, ImproperlyConfigured,
                         module_tables)
    from .strategy import create_engine
    from . import dialects

    __all__ = ['BaseModel', 'Model', 'Mapper', 'OdmSession', 'logger',
               'model_base', 'table_args', 'create_engine', 'dialects',
               'declared_attr', 'copy_models', 'move_models', 'get_models',
               'module_tables', 'ImproperlyConfigured']



#!/usr/bin/env python
import sys
import os
import json
import subprocess

from setuptools import setup, find_packages


def extend(params, package=None):
    if package:
        path = os.path.dirname(__file__)
        data = sh('%s %s package_info %s %s'
                  % (sys.executable, __file__, package, path))
        params.update(json.loads(data))

    return params


def read(name):
    filename = os.path.join(os.path.dirname(__file__), name)
    with open(filename) as fp:
        return fp.read()


def requirements(name):
    install_requires = []
    dependency_links = []

    for line in read(name).split('\n'):
        if line.startswith('-e '):
            link = line[3:].strip()
            if link == '.':
                continue
            dependency_links.append(link)
            line = link.split('=')[1]
        line = line.strip()
        if line:
            install_requires.append(line)

    return install_requires, dependency_links


def sh(command):
    return subprocess.Popen(command,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            shell=True,
                            universal_newlines=True).communicate()[0]


def package_info():
    package = sys.argv[2]
    if len(sys.argv) > 3:
        sys.path.append(sys.argv[3])
    os.environ['package_info'] = package
    pkg = __import__(package)
    print(json.dumps(dict(version=pkg.__version__,
                          description=pkg.__doc__)))


def run():
    requires, links = requirements('requirements.txt')

    meta = dict(
        name='pulsar-odm',
        author="Luca Sbardella",
        author_email="luca@quantmind.com",
        url="https://github.com/quantmind/pulsar-odm",
        zip_safe=False,
        license='BSD',
        long_description=read('README.rst'),
        packages=find_packages(include=('odm', 'odm.*')),
        setup_requires=['pulsar', 'wheel'],
        install_requires=requires,
        dependency_links=links,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Utilities'])

    setup(**extend(meta, 'odm'))


if __name__ == '__main__':
    command = sys.argv[1] if len(sys.argv) > 1 else None
    if command == 'package_info':
        package_info()
    elif command == 'agile':
        from agile.app import AgileManager
        AgileManager(description='Release manager for pulsar-odm',
                     argv=sys.argv[2:]).start()
    else:
        run()
